from gi.repository import Gtk

from gaphas.view import GtkView


class Zoom:
    def __init__(self, matrix):
        self.matrix = matrix
        self.x0 = 0
        self.y0 = 0
        self.sx = 1.0
        self.sy = 1.0

    def begin(self, x0, y0):
        self.x0 = x0
        self.y0 = y0
        self.sx = self.matrix[0]
        self.sy = self.matrix[3]

    def update(self, scale):
        if self.sx * scale < 0.2:
            scale = 0.2 / self.sx
        elif self.sx * scale > 20.0:
            scale = 20.0 / self.sx

        m = self.matrix
        sx = m[0]
        sy = m[3]
        ox = (m[4] - self.x0) / sx
        oy = (m[5] - self.y0) / sy
        dsx = self.sx * scale / sx
        dsy = self.sy * scale / sy
        m.translate(-ox, -oy)
        m.scale(dsx, dsy)
        m.translate(+ox, +oy)


def zoom_tool(view: GtkView) -> Gtk.GestureZoom:
    """Create a zoom tool as a Gtk.Gesture.

    Note: we need to keep a reference to this gesture, or else it will be destroyed.
    """
    zoom = Zoom(view.matrix)
    gesture = (
        Gtk.GestureZoom.new(view)
        if Gtk.get_major_version() == 3
        else Gtk.GestureZoom.new()
    )
    gesture.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
    gesture.connect("begin", on_begin, zoom)
    gesture.connect("scale-changed", on_scale_changed, zoom)
    return gesture


def on_begin(
    gesture: Gtk.GestureZoom,
    sequence: None,
    zoom: Zoom,
) -> None:
    _, x0, y0 = gesture.get_point(sequence)
    zoom.begin(x0, y0)


def on_scale_changed(gesture: Gtk.GestureZoom, scale: float, zoom: Zoom) -> None:
    zoom.update(scale)



from __future__ import annotations

from typing import Collection

import cairo

from gaphas.geometry import Rectangle
from gaphas.item import Item
from gaphas.painter.painter import ItemPainterType


class BoundingBoxPainter:
    """This specific case of an ItemPainter is used to calculate the bounding
    boxes (in cairo device coordinates) for the items.

    Deprecated since 3.2: Bounding boxes are calculated in the View
    directly. This class is not a pass-through for the provided
    item_painter.
    """

    def __init__(
        self,
        item_painter: ItemPainterType,
    ):
        self.item_painter = item_painter

    def paint_item(self, item: Item, cr: cairo.Context) -> None:
        self.item_painter.paint_item(item, cr)

    def paint(self, items: Collection[Item], cr: cairo.Context) -> None:
        """Draw the items, return the bounding boxes (in cairo device
        coordinates)."""
        self.item_painter.paint(items, cr)

    def bounding_box(self, items: Collection[Item], cr: cairo.Context) -> Rectangle:
        """Get the unified bounding box of the rendered items."""
        surface = cairo.RecordingSurface(cairo.Content.COLOR_ALPHA, None)  # type: ignore[arg-type]
        cr = cairo.Context(surface)
        self.paint(items, cr)
        return Rectangle(*surface.ink_extents())



"""Custom decorators."""
import functools
import inspect
import logging
import threading

from gi.repository import GLib

log = logging.getLogger(__name__)


class g_async:
    """Instead of calling the function, schedule an idle handler at a given
    priority. This requires the async'ed method to be called from within the
    GTK main loop. Otherwise the method is executed directly.

    If a function's first argument is "self", it's considered a method.

    Calling the async function from outside the gtk main loop will
    yield immediate execution.

    A function can also be a generator. The generator will be fully executed.
    If run in the main loop, an empty iterator will be returned.
    A generator is "single" by default. Because of the nature of generators
    the first invocation will run till completion.
    """

    def __init__(
        self,
        single: bool = False,
        timeout: int = 0,
        priority: int = GLib.PRIORITY_DEFAULT_IDLE,
    ) -> None:
        self.single = single
        self.timeout = timeout
        self.priority = priority

    def source(self, func):
        timeout = self.timeout
        s = GLib.Timeout(timeout) if timeout > 0 else GLib.Idle()
        s.set_callback(func)
        s.priority = self.priority
        return s

    def __call__(self, func):
        is_method = inspect.getfullargspec(func).args[:1] == ["self"]
        is_generator = inspect.isgeneratorfunction(func)
        source_attr = f"__g_async__{func.__name__}"

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # execute directly if we're not in the main loop
            if GLib.main_depth() == 0:
                return func(*args, **kwargs)
            elif is_generator:
                # We can only run one generator at a time
                holder = args[0] if is_method else func
                source = getattr(holder, source_attr, 0)
                if source:
                    return

                iterator = func(*args, **kwargs)

                def async_wrapper(*_args):
                    try:
                        next(iterator)
                    except Exception:
                        delattr(holder, source_attr)
                        return GLib.SOURCE_REMOVE
                    return GLib.SOURCE_CONTINUE

                source = self.source(async_wrapper)
                setattr(holder, source_attr, source)
                source.attach()
                return ()
            elif self.single:
                # Idle handlers should be registered per instance
                holder = args[0] if is_method else func
                source = getattr(holder, source_attr, 0)
                if source:
                    return

                def async_wrapper(*_args):
                    log.debug("async: %s %s %s", func, args, kwargs)
                    try:
                        func(*args, **kwargs)
                    finally:
                        delattr(holder, source_attr)
                    return GLib.SOURCE_REMOVE

                source = self.source(async_wrapper)
                setattr(holder, source_attr, source)
                source.attach()
            else:

                def async_wrapper(*_args):
                    log.debug("async: %s %s %s", func, args, kwargs)
                    func(*args, **kwargs)
                    return GLib.SOURCE_REMOVE

                self.source(async_wrapper).attach()

        return wrapper


def nonrecursive(func):
    """Enforce a function or method is not executed recursively:

    >>> class A(object):
    ...     @nonrecursive
    ...     def a(self, x=1):
    ...         print(x)
    ...         self.a(x+1)
    >>> A().a()
    1
    >>> A().a()
    1
    """
    m = threading.Lock()

    def wrapper(*args, **kwargs):
        """Decorate function with a mutex that prohibits recursive
        execution."""
        if m.acquire(False):
            try:
                return func(*args, **kwargs)
            finally:
                m.release()

    return wrapper



from typing import Callable, Optional

from gi.repository import Gtk

from gaphas.geometry import Rectangle
from gaphas.matrix import Matrix


class Scrolling:
    """Contains Gtk.Adjustment and related code."""

    def __init__(self, scrolling_updated: Callable[[Matrix], None]) -> None:
        self._scrolling_updated = scrolling_updated
        self.hadjustment: Optional[Gtk.Adjustment] = None
        self.vadjustment: Optional[Gtk.Adjustment] = None
        self.hscroll_policy: Optional[Gtk.ScrollablePolicy] = None
        self.vscroll_policy: Optional[Gtk.ScrollablePolicy] = None
        self._hadjustment_handler_id = 0
        self._vadjustment_handler_id = 0
        self._last_hvalue = 0
        self._last_vvalue = 0

    def get_property(self, prop):
        if prop.name == "hadjustment":
            return self.hadjustment
        elif prop.name == "vadjustment":
            return self.vadjustment
        elif prop.name == "hscroll-policy":
            return self.hscroll_policy
        elif prop.name == "vscroll-policy":
            return self.vscroll_policy
        else:
            raise AttributeError(f"Unknown property {prop.name}")

    def set_property(self, prop, value):
        if prop.name == "hadjustment":
            if value is not None:
                self.hadjustment = value
                self._hadjustment_handler_id = self.hadjustment.connect(
                    "value-changed", self.on_adjustment_changed
                )
                self._scrolling_updated(Matrix())
        elif prop.name == "vadjustment":
            if value is not None:
                self.vadjustment = value
                self._vadjustment_handler_id = self.vadjustment.connect(
                    "value-changed", self.on_adjustment_changed
                )
                self._scrolling_updated(Matrix())
        elif prop.name == "hscroll-policy":
            self.hscroll_policy = value
        elif prop.name == "vscroll-policy":
            self.vscroll_policy = value
        else:
            raise AttributeError(f"Unknown property {prop.name}")

    def update_adjustments(self, width, height, bounds):
        """Update scroll bar values (adjustments in GTK), and reset the scroll
        value to 0.

        The value will change when a scroll bar is moved.
        """
        # canvas limits (in view coordinates)
        c = Rectangle(*bounds)
        c.expand(min(width, height) / 2)
        u = c + Rectangle(0, 0, width, height)

        if self.hadjustment:
            self.hadjustment.set_value(0)
            self.hadjustment.set_lower(u.x)
            self.hadjustment.set_upper(u.x1)
            self.hadjustment.set_step_increment(width // 10)
            self.hadjustment.set_page_increment(width)
            self.hadjustment.set_page_size(width)
            self._last_hvalue = 0

        if self.vadjustment:
            self.vadjustment.set_value(0)
            self.vadjustment.set_lower(u.y)
            self.vadjustment.set_upper(u.y1)
            self.vadjustment.set_step_increment(height // 10)
            self.vadjustment.set_page_increment(height)
            self.vadjustment.set_page_size(height)
            self._last_vvalue = 0

    def on_adjustment_changed(self, adj):
        """Change the transformation matrix of the view to reflect the value of
        the x/y adjustment (scrollbar)."""
        value = adj.get_value()
        if value == 0:
            return

        m = Matrix()
        if adj is self.hadjustment:
            m.translate(self._last_hvalue - value, 0)
            self._last_hvalue = value
        elif adj is self.vadjustment:
            m.translate(0, self._last_vvalue - value)
            self._last_vvalue = value

        self._scrolling_updated(m)



# flake8: noqa F401
from gaphas.connector import (
    ConnectionSink,
    ConnectionSinkType,
    Connector,
    ElementConnectionSink,
    ItemConnectionSink,
    ItemConnector,
    LineConnector,
)



from gi.repository import Gdk, Gtk

from gaphas.tool.zoom import Zoom
from gaphas.view import GtkView


def scroll_tools(view: GtkView, speed: int = 10) -> Gtk.EventControllerScroll:
    return scroll_tool(view, speed), pan_tool(view)


def scroll_tool(view: GtkView, speed: int = 10) -> Gtk.EventControllerScroll:
    """Scroll tool recognized 2 finger scroll gestures."""
    ctrl = (
        Gtk.EventControllerScroll.new(
            view,
            Gtk.EventControllerScrollFlags.BOTH_AXES,
        )
        if Gtk.get_major_version() == 3
        else Gtk.EventControllerScroll.new(Gtk.EventControllerScrollFlags.BOTH_AXES)
    )
    ctrl.connect("scroll", on_scroll, speed)
    return ctrl


def on_scroll(controller, dx, dy, speed):
    view = controller.get_widget()

    modifiers = (
        Gtk.get_current_event_state()[1]
        if Gtk.get_major_version() == 3
        else controller.get_current_event_state()
    )

    if modifiers & Gdk.ModifierType.CONTROL_MASK:
        if Gtk.get_major_version() == 3:
            event = Gtk.get_current_event()
            x = event.x
            y = event.y
        else:
            # Workaround: Gtk.EventController.get_current_event() causes SEGFAULT
            view = controller.get_widget()
            x = view.get_width() / 2
            y = view.get_height() / 2
        zoom = Zoom(view.matrix)
        zoom.begin(x, y)

        zoom_factor = 0.1
        d = 1 - dy * zoom_factor
        zoom.update(d)
    elif modifiers & Gdk.ModifierType.SHIFT_MASK:
        view.hadjustment.set_value(dy * speed - view.hadjustment.get_value())
        view.vadjustment.set_value(dx * speed - view.vadjustment.get_value())
    else:
        view.hadjustment.set_value(dx * speed - view.hadjustment.get_value())
        view.vadjustment.set_value(dy * speed - view.vadjustment.get_value())


class PanState:
    def __init__(self):
        self.reset()

    def reset(self):
        self.pos = None


def pan_tool(view: GtkView) -> Gtk.GestureDrag:
    gesture = (
        Gtk.GestureDrag.new(view)
        if Gtk.get_major_version() == 3
        else Gtk.GestureDrag.new()
    )
    gesture.set_button(2)
    pan_state = PanState()
    gesture.connect("drag-begin", on_drag_begin, pan_state)
    gesture.connect("drag-update", on_drag_update, pan_state)
    return gesture


def on_drag_begin(gesture, _start_x, _start_y, pan_state):
    view = gesture.get_widget()
    m = view.matrix
    pan_state.pos = (m[4], m[5])
    gesture.set_state(Gtk.EventSequenceState.CLAIMED)


def on_drag_update(gesture, offset_x, offset_y, pan_state):
    view = gesture.get_widget()
    m = view.matrix
    x0, y0 = pan_state.pos
    m.translate(x0 + offset_x - m[4], y0 + offset_y - m[5])



"""Module implements guides when moving items and handles around."""
from __future__ import annotations

from dataclasses import dataclass
from functools import singledispatch
from itertools import chain
from typing import Iterable, SupportsFloat

from gaphas.canvas import all_children
from gaphas.connector import Handle
from gaphas.handlemove import ElementHandleMove, HandleMove, ItemHandleMove
from gaphas.item import Element, Item, Line
from gaphas.move import ItemMove, Move
from gaphas.types import Pos
from gaphas.view import GtkView


class ItemGuide:
    """Get edges on an item, on which we can align the items."""

    def __init__(self, item: Item, handle: Handle | None = None):
        self.item = item
        self.handle = handle

    def horizontal(self) -> Iterable[SupportsFloat]:
        """Return horizontal edges (on y axis) in item coordinates."""
        return ()

    def vertical(self) -> Iterable[SupportsFloat]:
        """Return vertical edges (on x axis) in item coordinates."""
        return ()


Guide = singledispatch(ItemGuide)


@Guide.register(Element)
class ElementGuide(ItemGuide):
    """Guide to align Element items."""

    def horizontal(self) -> Iterable[SupportsFloat]:
        element = self.item
        assert isinstance(element, Element)
        if self.handle in element.handles():
            return ()
        y = element.handles()[0].pos.y
        h = element.height
        return [y, y + h / 2, y + h]

    def vertical(self) -> Iterable[SupportsFloat]:
        element = self.item
        assert isinstance(element, Element)
        if self.handle in element.handles():
            return ()
        x = element.handles()[0].pos.x
        w = element.width
        return [x, x + w / 2, x + w]


@Guide.register(Line)
class LineGuide(ItemGuide):
    """Guide for orthogonal lines."""

    def horizontal(self) -> Iterable[SupportsFloat]:
        line = self.item
        assert isinstance(line, Line)
        if not line.orthogonal or self.handle not in line.handles():
            yield from (h.pos.y for h in line.handles() if h is not self.handle)

    def vertical(self) -> Iterable[SupportsFloat]:
        line = self.item
        assert isinstance(line, Line)
        if not line.orthogonal or self.handle not in line.handles():
            yield from (h.pos.x for h in line.handles() if h is not self.handle)


@dataclass(frozen=True)
class Guides:
    vertical: list[float]
    horizontal: list[float]


MARGIN = 4


def find_vertical_guides(view, handle, item_vedges, height, excluded_items, margin):
    items = (
        set(
            chain.from_iterable(
                view.get_items_in_rectangle((x - margin, 0, margin * 2, height))
                for x in item_vedges
            )
        )
        - excluded_items
    )
    guides = (Guide(item, handle) for item in items)
    i2v = view.get_matrix_i2v
    vedges = {
        i2v(g.item).transform_point(x, 0)[0] for g in guides for x in g.vertical()
    }
    dx, edges_x = find_closest(item_vedges, vedges)
    return dx, edges_x


def find_horizontal_guides(view, handle, item_hedges, width, excluded_items, margin):
    items = (
        set(
            chain.from_iterable(
                view.get_items_in_rectangle((0, y - margin, width, margin * 2))
                for y in item_hedges
            )
        )
        - excluded_items
    )
    guides = (Guide(item, handle) for item in items)
    i2v = view.get_matrix_i2v
    hedges = {
        i2v(g.item).transform_point(0, y)[1] for g in guides for y in g.horizontal()
    }
    dy, edges_y = find_closest(item_hedges, hedges)
    return dy, edges_y


def get_excluded_items(view, item):
    """Get a set of items excluded from guide calculation."""
    assert view.model

    excluded_items = set(all_children(view.model, item))
    excluded_items.add(item)
    excluded_items.update(view.selection.selected_items)
    return excluded_items


def get_view_dimensions(view):
    try:
        allocation = view.get_allocation()
    except AttributeError:
        return 0, 0
    return allocation.width, allocation.height


def find_closest(item_edges, edges, margin=MARGIN):
    delta = 0
    min_d = 1000
    closest = []
    for e in edges:
        for ie in item_edges:
            d = abs(e - ie)
            if d < min_d:
                min_d = d
                delta = e - ie
                closest = [e]
            elif d == min_d:
                closest.append(e)
    if min_d <= margin:
        return delta, closest
    else:
        return 0, ()


def update_guides(view, handle, pos, vedges, hedges, excluded_items=frozenset()):

    px, py = pos
    w, h = get_view_dimensions(view)
    dx, edges_x = find_vertical_guides(view, handle, vedges, h, excluded_items, MARGIN)
    dy, edges_y = find_horizontal_guides(
        view, handle, hedges, w, excluded_items, MARGIN
    )

    newpos = px + dx, py + dy

    view.guides = Guides(edges_x, edges_y)

    view.update_back_buffer()

    return newpos


def reset_guides(view):
    try:
        del view.guides
    except AttributeError:
        # No problem if guides do not exist.
        pass
    else:
        view.update_back_buffer()


class GuidedItemMoveMixin:
    """Move the item, lock position on any element that's located at the same
    location."""

    view: GtkView
    item: Item
    last_x: float
    last_y: float

    def move(self, pos: Pos) -> None:
        item = self.item
        view = self.view

        px, py = pos
        pdx, pdy = px - self.last_x, py - self.last_y

        transform = view.get_matrix_i2v(item).transform_point
        item_guide = Guide(item)
        item_vedges = [transform(x, 0)[0] + pdx for x in item_guide.vertical()]
        item_hedges = [transform(0, y)[1] + pdy for y in item_guide.horizontal()]

        excluded_items = get_excluded_items(view, item)
        newpos = update_guides(
            view, None, pos, item_vedges, item_hedges, excluded_items
        )

        # Call super class, with new position
        super().move(newpos)  # type: ignore[misc]

    def stop_move(self, pos: Pos) -> None:
        super().stop_move(pos)  # type: ignore[misc]
        reset_guides(self.view)


@Move.register(Element)
class GuidedItemMove(GuidedItemMoveMixin, ItemMove):
    pass


class GuidedItemHandleMoveMixin:
    """Move a handle and lock the position of other elements.

    Locks the position of another element that's located at the same
    position.
    """

    view: GtkView
    item: Item
    handle: Handle

    def move(self, pos: Pos) -> None:

        super().move(pos)  # type: ignore[misc]

        item = self.item
        view = self.view
        model = view.model
        assert model

        x, y = pos

        newpos = update_guides(view, self.handle, pos, (x,), (y,))

        self.handle.pos = view.get_matrix_v2i(item).transform_point(*newpos)
        model.request_update(item)

    def stop_move(self, pos: Pos) -> None:
        super().stop_move(pos)  # type: ignore[misc]
        reset_guides(self.view)


@HandleMove.register(Element)
class GuidedElementHandleMove(GuidedItemHandleMoveMixin, ElementHandleMove):
    pass


@HandleMove.register(Line)
class GuidedLineHandleMove(GuidedItemHandleMoveMixin, ItemHandleMove):
    pass


class GuidePainter:
    def __init__(self, view: GtkView):
        self.view = view

    def paint(self, _items, cr):
        try:
            guides = self.view.guides
        except AttributeError:
            return

        view = self.view
        allocation = view.get_allocation()
        w, h = allocation.width, allocation.height

        cr.save()
        try:
            cr.identity_matrix()
            cr.set_line_width(1)
            cr.set_source_rgba(0.0, 0.0, 1.0, 0.6)
            for g in guides.vertical:
                cr.move_to(g, 0)
                cr.line_to(g, h)
                cr.stroke()
            for g in guides.horizontal:
                cr.move_to(0, g)
                cr.line_to(w, g)
                cr.stroke()
        finally:
            cr.restore()
